# usbhotkey script to handle the Alphagrip AG5
# Copyright (C) 2007  Lars Krueger
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

# We assume a UK keymap here. For a US keymap exchange " and @

print "usbhotkey AG5 script starting\n"

# This script places all mappings in ruby tables which are indexed by some
# state variable and the USB scan code. The backside keys are mapped to a
# normalised id first.

# The basic idea of this script is to provide three color levels: red, green,
# and blue, which can be pressed together to acchieve mixed colours. Thus, 8
# additional shift levels can be implemented. These levels can also generate
# different keypresses with respect to the state of the shift key.
#
# All keys can send X11 keycode sequences of arbitrary length.

# Printscreen on the AG5 quits the program

##################################### Keymap ###################################
#==================== simple keys map from USB to X11 direct ===================
# Index 0: unshifted, index 1: shifted
SimpleKeys = []
SimpleKeys[0]={
  13 => [ WIN_Control_L ],
  25 => [ WIN_Alt_L ],
  27 => [ WIN_Alt_R ],
  29 => [ WIN_Control_R ],
  15 => [ WIN_Return ],
  44 => [ WIN_space ]}

SimpleKeys[1]={
  13 => [ WIN_Control_L ],
  25 => [ WIN_Alt_L ],
  27 => [ WIN_Alt_R ],
  29 => [ WIN_Control_R ],
  15 => [ WIN_Tab ],
  42 => [ WIN_BackSpace ],
  44 => [ WIN_Escape ]}

#======================== Set or release the colorshift ========================
ColorKeys = {
  6  => [0, 2],
  28 => [0, 4],
  14 => [0, 1],
  40 => [1, 4],
  42 => [1, 2],
  43 => [1, 1]}

#=========================== USB code to backside id ===========================
BacksideId={
  USB_g         => 0 ,
  USB_r         => 1 ,
  USB_f         => 2 ,
  USB_t         => 3 ,
  USB_d         => 4 ,
  USB_e         => 5 ,
  USB_w         => 6 ,
  USB_s         => 7 ,
  USB_q         => 8 ,
  USB_a         => 9 ,
  USB_m         => 10,
  USB_h         => 11,
  USB_n         => 12,
  USB_u         => 13,
  USB_i         => 14,
  USB_Comma     => 15,
  USB_o         => 16,
  USB_Period    => 17,
  USB_p         => 18,
  USB_b         => 19,
  }

#================================ Normal keymap ================================
# Dimension 1 by ORed colorshift values
# Dimension 2 shift state
# Dimension 3 by backside ID
KeyMap=[[], []]

# g|r|f|t|d|e|w|s|q|a|m|h|n|u|i|l|o|c|p|b"
KeyMap[0][0] = [ 
  [ WIN_G],
  [ WIN_R],
  [ WIN_F],
  [ WIN_T],
  [ WIN_D],
  [ WIN_E],
  [ WIN_W],
  [ WIN_S],
  [ WIN_Q],
  [ WIN_A],
  [ WIN_M],
  [ WIN_H],
  [ WIN_N],
  [ WIN_U],
  [ WIN_I],
  [ WIN_L],
  [ WIN_O],
  [ WIN_C],
  [ WIN_P],
  [ WIN_B]]

# {up}|{down}|{left}|{right}|{tab}|{space}|:|{del}|@|{esc}|{pgup}|{pgdn}|{home}|{end}|{enter}|{bs}|.|;|,|{printscreen}"
KeyMap[0][1] = [ 
  [ WIN_Up],
  [ WIN_Down],
  [ WIN_Left],
  [ WIN_Right],
  [ WIN_Tab],
  [ WIN_space],
  [ WIN_Shift_L, WIN_semicolon],
  [ WIN_Delete],
  [ WIN_Shift_L, WIN_apostrophe],
  [ WIN_Escape],
  [ WIN_Page_Up],
  [ WIN_Page_Down],
  [ WIN_Home],
  [ WIN_End],
  [ WIN_Return],
  [ WIN_BackSpace],
  [ ],
  [ WIN_semicolon],
  [ ],
  [ WIN_Print]]

# {#}|2|,|1|{+}|3|-|4|=|5|$|7|6|.|8|*|9|;|0|/"
KeyMap[0][2] = [ 
  [ WIN_numbersign ],
  [ WIN_2 ],
  [ WIN_period ],
  [ WIN_1 ],
  [ WIN_Shift_L, WIN_equal ],
  [ WIN_3 ],
  [ WIN_minus ],
  [ WIN_4 ],
  [ WIN_equal ],
  [ WIN_5 ],
  [ WIN_Shift_L, WIN_4 ],
  [ WIN_7 ],
  [ WIN_6 ],
  [ WIN_comma ],
  [ WIN_8 ],
  [ WIN_Shift_L, WIN_8 ],
  [ WIN_9 ],
  [ WIN_semicolon ],
  [ WIN_0 ],
  [ WIN_slash ]]

# | | |j||v| |y| || | |k||x| |z|| | "
KeyMap[0][3] = [ 
  [  ],
  [  ],
  [  ],
  [ WIN_J ],
  [  ],
  [ WIN_V ],
  [  ],
  [ WIN_Y ],
  [  ],
  [  ],
  [  ],
  [  ],
  [ WIN_K ],
  [  ],
  [ WIN_X ],
  [  ],
  [ WIN_Z ],
  [  ],
  [  ],
  [  ]]

# ~,%,"",',(,),{{},{}},[,],{^},_,?,!,&,|,<,>,``,\"
KeyMap[0][4] = [ 
  [ WIN_Shift_L, WIN_numbersign ],
  [ WIN_Shift_L, WIN_5 ],
  [ WIN_Shift_L, WIN_2 ],
  [ WIN_apostrophe],
  [ WIN_Shift_L, WIN_9 ],
  [ WIN_Shift_L, WIN_0 ],
  [ WIN_Shift_L, WIN_bracketleft ],
  [ WIN_Shift_L, WIN_bracketright ],
  [ WIN_bracketleft  ],
  [ WIN_bracketright ],
  [ WIN_Shift_L, WIN_6 ],
  [ WIN_Shift_L, WIN_minus ],
  [ WIN_Shift_L, WIN_slash ],
  [ WIN_Shift_L, WIN_1 ],
  [ WIN_Shift_L, WIN_7 ],
  [ WIN_Shift_L, WIN_backslash ],
  [ WIN_Shift_L, WIN_comma ],
  [ WIN_Shift_L, WIN_period ],
  [ WIN_quoteleft ],
  [ WIN_backslash ]]

# | | | | | | | | | | | | | | | | | | | "
KeyMap[0][5] = [ 
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ]]

# {F11}|{F2}|{F1}| |{F3}| |{F4}| |{F5}| |{F12}|{F7}| |{F6}| |{F8}| |{F9}| |{F10}"
KeyMap[0][6] = [ 
  [ WIN_F11 ],
  [ WIN_F2 ],
  [ WIN_F1 ],
  [  ],
  [ WIN_F3 ],
  [  ],
  [ WIN_F4 ],
  [  ],
  [ WIN_F5 ],
  [  ],
  [ WIN_F12 ],
  [ WIN_F7 ],
  [  ],
  [ WIN_F6 ],
  [  ],
  [ WIN_F8 ],
  [  ],
  [ WIN_F9 ],
  [  ],
  [ WIN_F10 ]]


# | | | | | | | | | | | | | | | | | | | "
KeyMap[0][7] = [ 
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ]]


# g|r|f|t|d|e|w|s|q|a|m|h|n|u|i|l|o|c|p|b"
KeyMap[1][0] = [ 
  [ WIN_Shift_L, WIN_G],
  [ WIN_Shift_L, WIN_R],
  [ WIN_Shift_L, WIN_F],
  [ WIN_Shift_L, WIN_T],
  [ WIN_Shift_L, WIN_D],
  [ WIN_Shift_L, WIN_E],
  [ WIN_Shift_L, WIN_W],
  [ WIN_Shift_L, WIN_S],
  [ WIN_Shift_L, WIN_Q],
  [ WIN_Shift_L, WIN_A],
  [ WIN_Shift_L, WIN_M],
  [ WIN_Shift_L, WIN_H],
  [ WIN_Shift_L, WIN_N],
  [ WIN_Shift_L, WIN_U],
  [ WIN_Shift_L, WIN_I],
  [ WIN_Shift_L, WIN_L],
  [ WIN_Shift_L, WIN_O],
  [ WIN_Shift_L, WIN_C],
  [ WIN_Shift_L, WIN_P],
  [ WIN_Shift_L, WIN_B]]


# {up}|{down}|{left}|{right}|{tab}|{space}|:|{del}|@|{esc}|{pgup}|{pgdn}|{home}|{end}|{enter}|{bs}|.|;|,|{printscreen}"
KeyMap[1][1] = [ 
  [ WIN_Shift_L, WIN_Up],
  [ WIN_Shift_L, WIN_Down],
  [ WIN_Shift_L, WIN_Left],
  [ WIN_Shift_L, WIN_Right],
  [ WIN_Shift_L, WIN_Tab],
  [ WIN_Shift_L, WIN_space],
  [ WIN_Shift_L, WIN_colon],
  [ WIN_Shift_L, WIN_Delete],
  [ WIN_Shift_L, WIN_at],
  [ WIN_Shift_L, WIN_Escape],
  [ WIN_Shift_L, WIN_Page_Up],
  [ WIN_Shift_L, WIN_Page_Down],
  [ WIN_Shift_L, WIN_Home],
  [ WIN_Shift_L, WIN_End],
  [ WIN_Shift_L, WIN_Return],
  [ WIN_Shift_L, WIN_BackSpace],
  [ WIN_Shift_L, ],
  [ WIN_Shift_L, WIN_semicolon],
  [ WIN_Shift_L, ],
  [ WIN_Shift_L, WIN_Print]]


# {#}|2|,|1|{+}|3|-|4|=|5|$|7|6|.|8|*|9|;|0|/"
KeyMap[1][2] = [ 
  [ WIN_numbersign ],
  [ WIN_2 ],
  [ WIN_comma ],
  [ WIN_1 ],
  [ WIN_Shift_L, WIN_equal ],
  [ WIN_3 ],
  [ WIN_minus ],
  [ WIN_4 ],
  [ WIN_equal ],
  [ WIN_5 ],
  [ WIN_Shift_L, WIN_4 ],
  [ WIN_7 ],
  [ WIN_6 ],
  [ WIN_period ],
  [ WIN_8 ],
  [ WIN_Shift_L, WIN_8 ],
  [ WIN_9 ],
  [ WIN_semicolon ],
  [ WIN_0 ],
  [ WIN_slash ]]

#  | | |j||v| |y| || | |k||x| |z|| | "
KeyMap[1][3] = [ 
  [  ],
  [  ],
  [  ],
  [ WIN_Shift_L, WIN_J ],
  [  ],
  [ WIN_Shift_L, WIN_V ],
  [  ],
  [ WIN_Shift_L, WIN_Y ],
  [  ],
  [  ],
  [  ],
  [  ],
  [ WIN_Shift_L, WIN_K ],
  [  ],
  [ WIN_Shift_L, WIN_X ],
  [  ],
  [ WIN_Shift_L, WIN_Z ],
  [  ],
  [  ],
  [  ]]

# ~,%,"",',(,),{{},{}},[,],{^},_,?,!,&,|,<,>,``,\"
KeyMap[1][4] = [ 
  [ WIN_Shift_L, WIN_numbersign ],
  [ WIN_Shift_L, WIN_5 ],
  [ WIN_Shift_L, WIN_2 ],
  [ WIN_apostrophe],
  [ WIN_Shift_L, WIN_9 ],
  [ WIN_Shift_L, WIN_0 ],
  [ WIN_Shift_L, WIN_bracketleft ],
  [ WIN_Shift_L, WIN_bracketright ],
  [ WIN_bracketleft  ],
  [ WIN_bracketright ],
  [ WIN_Shift_L, WIN_6 ],
  [ WIN_Shift_L, WIN_minus ],
  [ WIN_Shift_L, WIN_slash ],
  [ WIN_Shift_L, WIN_1 ],
  [ WIN_Shift_L, WIN_7 ],
  [ WIN_Shift_L, WIN_backslash ],
  [ WIN_Shift_L, WIN_comma ],
  [ WIN_Shift_L, WIN_period ],
  [ WIN_quoteleft ],
  [ WIN_backslash ]]

#  | | | | | | | | | | | | | | | | | | | "
KeyMap[1][5] = [ 
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ]]

# {F11}|{F2}|{F1}| |{F3}| |{F4}| |{F5}| |{F12}|{F7}| |{F6}| |{F8}| |{F9}| |{F10}"
KeyMap[1][6] = [ 
  [ WIN_Shift_L, WIN_F11 ],
  [ WIN_Shift_L, WIN_F2 ],
  [ WIN_Shift_L, WIN_F1 ],
  [  ],
  [ WIN_Shift_L, WIN_F3 ],
  [  ],
  [ WIN_Shift_L, WIN_F4 ],
  [  ],
  [ WIN_Shift_L, WIN_F5 ],
  [  ],
  [ WIN_Shift_L, WIN_F12 ],
  [ WIN_Shift_L, WIN_F7 ],
  [  ],
  [ WIN_Shift_L, WIN_F6 ],
  [  ],
  [ WIN_Shift_L, WIN_F8 ],
  [  ],
  [ WIN_Shift_L, WIN_F9 ],
  [  ],
  [ WIN_Shift_L, WIN_F10 ]]

# | | | | | | | | | | | | | | | | | | | "
KeyMap[1][7] = [ 
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ]]

##################################### state ####################################
# State of the colour shift keys. Index 0 is for the left hand side keys.
$colorshift=[0, 0]

# State of the normal shift key, bit 0 is for left shift.
$shift=0

##################################### action ###################################

# Send the key presses in the list.
def process_keydn_list( list)
  for i in 0...list.length
    UHK_keydn( list[i])
  end
end

# Send the key releases in the list in reverse order.
def process_keyup_list( list)
  n=list.length
  n1=n-1
  for i in 0...n
    UHK_keyup( list[n1-i])
  end
end

# Handle the start and stop marks.
def UHK_USB_mark( code, millisecs)
  # print "Mark: #{code}, #{millisecs}\n"
end

# Handle key press events.
def UHK_USB_keydn( code)
  # print "Down #{code}\n"
 
  if code == USB_PrintScreen
    exit(1)
  end

  # Check if a shift key is pressed.
  if code == USB_SHIFT_L
    $shift=$shift | 1
    return
  end
  if code == USB_SHIFT_R
    $shift=$shift | 2
    return
  end

  # Flatten the shift state into a single bit
  shiftInd=0
  if not ($shift == 0)
    shiftInd=1
  end

  # Simple keys can be mapped directly from USB code to X11 code. There are
  # some keys that are to be ignored.
  if SimpleKeys[shiftInd].has_key?( code)
    # print " simple  -> #{SimpleKeys[$shift][code]}\n"
    process_keydn_list( SimpleKeys[shiftInd][code])
    return
  end
  #  Set the resective bits of the colorshift variables and update the overlay
  #  image.
  if ColorKeys.has_key?(code)
    $colorshift[ColorKeys[code][0]] = $colorshift[ColorKeys[code][0]] | ColorKeys[code][1]
    # print "  color #{$colorshift}\n"
    cs=$colorshift[0] | $colorshift[1]
    return
  end

  # Map the backside keys to their IDs and map those IDs to the X11 codes.
  # First handle the special case of shift-, send as ;. Turn it into shift-,
  if code==USB_Semicolon
    # print "Shift: #{shiftInd}\n"
    if (shiftInd==0) 
      code=USB_Comma
      shiftInd=1
    else
      code=USB_Period
    end
  end

  if BacksideId.has_key?(code)
    id=BacksideId[code]
    cs=$colorshift[0] | $colorshift[1]
    process_keydn_list( KeyMap[shiftInd][cs][id])
  end
end

def UHK_USB_keyup( code)
  # Shift
  # print "Up #{code}\n"
  if code == USB_SHIFT_L
    $shift=$shift & 2
    return
  end
  if code == USB_SHIFT_R
    $shift=$shift & 1
    return
  end

  shiftInd=0
  if not ($shift == 0)
    shiftInd=1
  end

  # Simple keys
  if SimpleKeys[shiftInd].has_key?( code)
    # print " simple  -> #{SimpleKeys[$shift][code]}\n"
    process_keyup_list( SimpleKeys[shiftInd][code]);
    return
  end
  # Colorshift
  if ColorKeys.has_key?(code)
    $colorshift[ColorKeys[code][0]] = $colorshift[ColorKeys[code][0]] & ( ColorKeys[code][1] ^ 7)
    # print "  color #{$colorshift}\n"
    cs=$colorshift[0] | $colorshift[1]
    return
  end
  # Backside keys
  # First handle the special cases
  if code==USB_Semicolon
    if (shiftInd==0) 
      code=USB_Comma
      shiftInd=1
    else
      code=USB_Period
    end
  end
  if BacksideId.has_key?(code)
    id=BacksideId[code]
    cs=$colorshift[0] | $colorshift[1]
    # print " back -> #{ks}\n"
    process_keyup_list( KeyMap[shiftInd][cs][id])
  end
end

# Connect to the alphagrip if present.
UHK_connect( 0x510, 0x100b, 0x81)

print "usbhotkey AG5 script started\n"

